/**
 * LICENCIA LGPL:
 * 
 * Esta librera es Software Libre; Usted puede redistribuirla y/o modificarla
 * bajo los trminos de la GNU Lesser General Public License (LGPL) tal y como 
 * ha sido publicada por la Free Software Foundation; o bien la versin 2.1 de 
 * la Licencia, o (a su eleccin) cualquier versin posterior.
 * 
 * Esta librera se distribuye con la esperanza de que sea til, pero SIN 
 * NINGUNA GARANTA; tampoco las implcitas garantas de MERCANTILIDAD o 
 * ADECUACIN A UN PROPSITO PARTICULAR. Consulte la GNU Lesser General Public 
 * License (LGPL) para ms detalles
 * 
 * Usted debe recibir una copia de la GNU Lesser General Public License (LGPL) 
 * junto con esta librera; si no es as, escriba a la Free Software Foundation 
 * Inc. 51 Franklin Street, 5 Piso, Boston, MA 02110-1301, USA o consulte
 * <http://www.gnu.org/licenses/>.
 *
 * Copyright 2011 Agencia de Tecnologa y Certificacin Electrnica
 */
package es.accv.arangi.base.device.model;

import java.util.HashMap;

import org.apache.log4j.Logger;

/**
 * Los distintos modelos de dispositivos registran en esta clase su mdulo asociado.
 * Accediendo al mtodo getInstance se obtienen estos mdulos con los que se pueden
 * abrir los dispositivos.
 * 
 * @author <a href="mailto:jgutierrez@accv.es">Jos M Gutirrez</a>
 */
public class CardModuleFactory {

	/**
	 * Logger de la clase
	 */
	static Logger logger = Logger.getLogger(CardModuleFactory.class);
	
	static HashMap hmModules;
	
	/**
	 * Registra un nuevo mdulo para tratar un tipo de dispositivo
	 * 
	 * @param deviceType Nombre del tipo de dispositivo
	 * @param cardModule Mdulo 
	 */
	public static void register (String deviceType, Pkcs11Manufacturer cardModule) {
		getModules().put (deviceType, cardModule);
	}
	
	/**
	 * Obtiene un mdulo para tratar con el tipo de dispositivo pasado como
	 * parmetro.
	 * 
	 * @param deviceType Tipo de dispositivo
	 * @return Modulo o null si no existe ningn mdulo registrado para el
	 * tipo de dispositivo
	 */
	public static Pkcs11Manufacturer getInstance (String deviceType) {
		return (Pkcs11Manufacturer) getModules().get(deviceType);
	}

	/*
	 * Obtiene el mapa de mdulos
	 */
	private static HashMap getModules() {
		if (hmModules == null) {
			hmModules = new HashMap ();
		}
		return hmModules;
	}
}
